from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant)
from qgis.core import (
  QgsProcessingParameterFeatureSource,
  QgsProcessing,
  QgsProcessingParameterPointCloudAttribute,
  QgsProcessingParameterDistance,
  QgsProcessingParameterFeatureSink,
  QgsFeatureRequest,
  QgsFeature,
  QgsField,
  QgsWkbTypes,
  QgsVectorLayer,
  QgsProcessingParameterField,
  QgsProcessingParameterPointCloudLayer,
  QgsFields,
  QgsProcessingParameterBoolean,
  QgsProcessingParameterNumber,
  QgsProcessingParameterFolderDestination,
  QgsVectorFileWriter,
  QgsCoordinateTransformContext
  )
from qgis import processing

from .algabstract import algabstract
import math
import os
import json

class crosssection(algabstract):
  
  # UIs
  PARAMETERS = {            
    "POINTCLOUD": {
      "ui_func": QgsProcessingParameterPointCloudLayer,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("crosssection","Point cloud layer")
      }
    },
    "POINTCLOUD_ATTRIBUTES": {
      "ui_func": QgsProcessingParameterPointCloudAttribute,
      "ui_args":{
        "optional": True,
        "description": QT_TRANSLATE_NOOP("crosssection","Attributes of point cloud layer"),
        "parentLayerParameterName": "POINTCLOUD",
        "allowMultiple": True
      }
    },
    "LINE": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("crosssection","Line layer"),
        "types": [QgsProcessing.TypeVectorLine]
      }
    },
    "LINE_ORIGIN_X": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "optional": True,
        "description": QT_TRANSLATE_NOOP("crosssection","X value of the origin of line layer"),
        "defaultValue": "pnt_orig_x",
        "parentLayerParameterName": "LINE"
      }
    },
    "LINE_ORIGIN_Y": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "optional": True,
        "description": QT_TRANSLATE_NOOP("crosssection","Y value of the origin of line layer"),
        "defaultValue": "pnt_orig_y",
        "parentLayerParameterName": "LINE"
      }
    },
    "LINE_PK": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "optional": True,
        "description": QT_TRANSLATE_NOOP("crosssection","Primary key of line layer"),
        "defaultValue": "transect_id",
        "parentLayerParameterName": "LINE"
      }
    },
    "LINE_OTHER_ATTRIBUTES": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "optional": True,
        "description": QT_TRANSLATE_NOOP("crosssection","Attributes of line layer"),
        "parentLayerParameterName": "LINE",
        "allowMultiple": True
      }
    },
    "BUFFER": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("crosssection","Width of the line buffer"),
        "defaultValue": 5.0,
        "parentParameterName": "LINE"
      }
    },      
    "N_MAX": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("crosssection","Maximum number of outputs"),
        "type": QgsProcessingParameterNumber.Integer,
        "minValue": 1, "defaultValue": 100, "maxValue": 99999        
      }
    },      
    "OUTPUT_AS_FILES": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("crosssection","Output as files (or layers)"),
        "defaultValue": False
      }
    },
    "OVERWRITE_FILES": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("crosssection","Overwrite files?"),
        "defaultValue": False
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("crosssection","Crosssection points")
      }
    },
    "OUTPUT_FOLDER": {
      "ui_func": QgsProcessingParameterFolderDestination,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("crosssectionfiles","Folder to output")
      }
    }
  }
  
  FIELDS_ADD = {
    "X_ORIGIN": {"TYPE": QVariant.Double, "DEFAULT_VALUE": 0.0},
    "Y_ORIGIN": {"TYPE": QVariant.Double, "DEFAULT_VALUE": 0.0},
    "ANGLE_DEG":{"TYPE": QVariant.Double, "DEFAULT_VALUE": 0.0},    
    "X_PROJ":   {"TYPE": QVariant.Double, "DEFAULT_VALUE": 0.0},    
    "DIST":     {"TYPE": QVariant.Double, "DEFAULT_VALUE": 0.0},
    "Z_PROJ":   {"TYPE": QVariant.Double, "DEFAULT_VALUE": 0.0}
  }
    
  
  # initialization of the algorithm
  def initAlgorithm(self, config):   
    self.initParameters()
  
  def outputFilePath(self, output_dir, num):
    return os.path.join(self.output_dir, f"PtCl-Crs{num:05}.geojson")
  
  # execution of the algorithm
  def processAlgorithm(self, parameters, context, feedback):    
    # import ptvsd
    # ptvsd.debug_this_thread()
    line = self.parameterAsSource(parameters, "LINE", context).materialize(QgsFeatureRequest(), feedback)
    target_crs = line.sourceCrs()
    line_orig_x = self.parameterAsString(parameters, "LINE_ORIGIN_X", context)
    line_orig_y = self.parameterAsString(parameters, "LINE_ORIGIN_Y", context)
    
    # prepare the fields
    output_flds = QgsFields()
    
    # add the fields from the point cloud
    pc_flds = self.parameterAsPointCloudLayer(parameters, "POINTCLOUD", context).attributes().toFields()
    for fld in self.parameterAsStrings(parameters, "POINTCLOUD_ATTRIBUTES", context):
      output_flds.append(pc_flds[fld])
    
    # add the fields from the line
    for fld in line.fields():
      if fld.name() in self.parameterAsStrings(parameters, "LINE_OTHER_ATTRIBUTES", context):
        if fld.name() != line_orig_x and fld.name() != line_orig_y:
          output_flds.append(fld)
    
    # add the fields for specifying the locations of the points
    for k, v in self.FIELDS_ADD.items():
      output_flds.append(QgsField(k, v["TYPE"]))
    
    
    # set the sink
    if self.parameterAsBool(parameters, "OUTPUT_AS_FILES", context):
      output_mode = "file"
      folder_path = self.parameterAsString(parameters, "OUTPUT_FOLDER", context)
      if not os.path.exists(folder_path):
        os.makedirs(folder_path)
      dest_id = None
      
      save_options = QgsVectorFileWriter.SaveVectorOptions()
      save_options.driverName = "GeoJSON"
      
      def init_vlyr(flds):
        vl = QgsVectorLayer("Point?crs=" + target_crs.authid(), "temp", "memory")
        vl.dataProvider().addAttributes(flds)
        vl.updateFields()
        return vl
      
    else:
      output_mode = "layer"
      (sink, dest_id) = self.parameterAsSink(
        parameters, "OUTPUT", context,
        output_flds, QgsWkbTypes.Point, target_crs
      )
    
    
    # prepare the line buffer
    line_buffer = processing.run(
      "native:buffer",
      {
        "INPUT": self.parameterAsSource(parameters, "LINE", context).materialize(QgsFeatureRequest(), feedback),
        "DISTANCE": self.parameterAsDouble(parameters, "BUFFER", context),
        "SEGMENTS": 5,
        "DISSOLVE": False,
        "END_CAP_STYLE": 1,
        "JOIN_STYLE": 0,
        "MITER_LIMIT": 2,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm = True
    )["OUTPUT"]
    
    line_buffer = context.getMapLayer(line_buffer)
    
    
    # for each line and the buffer, clip the point cloud layer
    for ft_line, ft_buf in zip(line.getFeatures(), line_buffer.getFeatures()):
      
      # get the primary key of the line      
      ft_line_id = ft_line[self.parameterAsString(parameters, "LINE_PK", context)]
      
      
      # save the vector layer, if the output mode is file 
      if output_mode == "file":
        file_path = os.path.join(folder_path, f"PtCl-Crs{ft_line_id:05}.geojson")
        if os.path.exists(file_path):
          if self.parameterAsBool(parameters, "OVERWRITE_FILES", context):
            os.remove(file_path)
          else:
            continue
      
      # get the origin of the line, the angle of the line, and the unit vector of the line
      # which are used to calculate the projection of the point cloud
      pnts_on_line = ft_line.geometry().asPolyline()
      if line_orig_x is not None and line_orig_y is not None:
        pnt_origin = [ft_line[line_orig_x], ft_line[line_orig_y]]
      else:
        pnt_origin = [pnts_on_line[0].x(), pnts_on_line[0].y()]
      angle_vector = [pnts_on_line[-1].x() - pnts_on_line[0].x(), pnts_on_line[-1].y() - pnts_on_line[0].y()]
      angle_deg = 450 - math.atan2(angle_vector[1], angle_vector[0]) * 180 / math.pi # note! it is north clockwise
      angle_deg = angle_deg - 360.0 if angle_deg > 360.0 else angle_deg
      length = (angle_vector[0] ** 2 + angle_vector[1] ** 2) ** 0.5
      unit_vector = [angle_vector[0] / length, angle_vector[1] / length]
      
      # prepare the buffer for clipping the point cloud layer
      clip_buffer = QgsVectorLayer(f"Polygon?crs={target_crs.authid()}", "temp", "memory")
      clip_buffer.dataProvider().addFeatures([ft_buf])
      
      # clip the point cloud layer
      # the output is either a point cloud or a virtual point cloud
      pc_or_vpc_in_buffer = processing.run(
        "pdal:clip",
        {
          "INPUT": self.parameterAsPointCloudLayer(parameters, "POINTCLOUD", context),
          "OVERLAY": clip_buffer,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        feedback = feedback,
        is_child_algorithm = True
      )["OUTPUT"]
      
      # if the output is virtual point cloud, set the actual point clouds as the inputs for the next step
      pc_in_buffer = []
      if os.path.splitext(pc_or_vpc_in_buffer)[1] == ".vpc":
        with open(pc_or_vpc_in_buffer) as vpc:
          vpc_contents = json.load(vpc)
        for pc in vpc_contents["features"]:
          rel_path = pc["assets"]["data"]["href"]
          abs_path = os.path.join(os.path.normpath(os.path.dirname(pc_or_vpc_in_buffer)), rel_path.replace("./", ""))
          pc_in_buffer.append(abs_path)
      else:
        pc_in_buffer.append(pc_in_buffer)
      
      # for each point cloud in the buffer, export the vector layer
      vecs_from_pc = []
      for pc in pc_in_buffer:
        single_vec = processing.run(
          "pdal:exportvector",
          {
            "INPUT": pc,
            "ATTRIBUTE": self.parameterAsStrings(parameters, "POINTCLOUD_ATTRIBUTES", context),
            "OUTPUT": "TEMPORARY_OUTPUT"
          },
          context = context,
          feedback = feedback,
          is_child_algorithm = True
        )["OUTPUT"]
        vecs_from_pc.append(single_vec)
      
      # merge the vector layers, if necessary
      if len(vecs_from_pc) == 1:
        vec_from_pc = QgsVectorLayer(vecs_from_pc[0])
      else:
        vec_from_pc = processing.run(
          "native:mergevectorlayers",
          {
            "LAYERS": vecs_from_pc,
            "CRS": target_crs,
            "OUTPUT": "TEMPORARY_OUTPUT"
          },
          context = context,
          feedback = feedback,
          is_child_algorithm = True
        )["OUTPUT"]
        vec_from_pc = context.getMapLayer(vec_from_pc)
      
      
      if output_mode == "file":
        vec_layer = init_vlyr(output_flds)
        vec_layer_pr = vec_layer.dataProvider()
      
      # make the features and add them to the sink
      for ft_pnt in vec_from_pc.getFeatures():
        pc_x = ft_pnt.geometry().get().x()
        pc_y = ft_pnt.geometry().get().y()
        pc_z = ft_pnt.geometry().get().z()
        x_proj = (pc_x - pnt_origin[0]) * unit_vector[0] + (pc_y - pnt_origin[1]) * unit_vector[1]
        dist_sq = (pc_x - pnt_origin[0]) * (pc_x - pnt_origin[0]) + (pc_y - pnt_origin[1]) * (pc_y - pnt_origin[1]) - x_proj * x_proj
        dist = math.sqrt(dist_sq) if dist_sq > 0.0 else 0.0
        
        # set output geometry
        ft_output = QgsFeature(output_flds)
        ft_output.setGeometry(ft_pnt.geometry())
        
        # set output fields from point cloud
        for fld in ft_pnt.fields():
          if fld.name() in self.parameterAsStrings(parameters, "POINTCLOUD_ATTRIBUTES", context):
            ft_output[fld.name()] = ft_pnt[fld.name()]
        
        # set output fields from line
        for fld in ft_line.fields():
          if fld.name() in self.parameterAsStrings(parameters, "LINE_OTHER_ATTRIBUTES", context) and \
            fld.name() != line_orig_x and fld.name() != line_orig_y:
              ft_output[fld.name()] = ft_line[fld.name()]
        
        # set output fields for specifying the locations of the points
        ft_output["X_ORIGIN"] = pnt_origin[0]
        ft_output["Y_ORIGIN"] = pnt_origin[1]
        ft_output["ANGLE_DEG"] = angle_deg
        ft_output["X_PROJ"] = x_proj
        ft_output["DIST"] = dist
        ft_output["Z_PROJ"] = pc_z
        
        # add the feature to the sink
        if output_mode == "file":
          vec_layer_pr.addFeatures([ft_output])
        else:
          sink.addFeature(ft_output)
      
      # save the vector layer, if the output mode is file 
      if output_mode == "file":
        try:
          QgsVectorFileWriter.writeAsVectorFormatV3(vec_layer, file_path, QgsCoordinateTransformContext(), save_options)
          # file_path = os.path.join(folder_path, f"PtCl-Crs{ft_line_id:05}.geojson")
          # if os.path.exists(file_path):
          #   raise Exception()
          
        except:
          feedback.pushWarning(f"Failed to save the file: {file_path}")
          # file_path = os.path.join(folder_path, f"Autoname-PtCl-Crs{ft_line.id():05}.geojson")
          # QgsVectorFileWriter.writeAsVectorFormatV3(vec_layer, file_path, QgsCoordinateTransformContext(), save_options)
        
    
    if output_mode == "file":
      return {"OUTPUT": folder_path}
    else:
      return {"OUTPUT": dest_id}
  
  def displayName(self):
    return self.tr("Crosssection")

  def group(self):
    return self.tr('Tools')

  def groupId(self):
    return 'tools'

  def createInstance(self):
    return crosssection()
