from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP)
from qgis.core import (
  QgsProcessingContext,
  QgsProcessingFeedback,
  QgsProcessingUtils,
  QgsCoordinateReferenceSystem
  )

from .algabstract import algabstract
from qgis import processing

import os
import shutil
import glob
import os
import requests
import concurrent

class fetchabstract(algabstract):
  
  TENGUN_EXTENSION = ["las", "laz"]
  ARCHIVE_EXTENSION = ["zip", "tar", "gztar", "bztar", "zxtar"]
  OUTPUT_FOLDER = ""
  
  def setWorkAndOutputDir(self, working_dir:str = None, output_dir:str = None) -> None:
    # set working directory
    if working_dir is None:
      working_dir = os.path.normpath(os.path.dirname(QgsProcessingUtils.generateTempFilename("")))
    else:
      working_dir = os.path.normpath(working_dir)

    # set output directory
    if output_dir is None:
      output_dir = os.path.normpath(os.path.dirname(QgsProcessingUtils.generateTempFilename("")))
    else:
      output_dir = os.path.normpath(output_dir)
    
    return working_dir, output_dir
  
  # download a tengun file, unzip it, and assign projection
  def fetchSingleTengun(self, session, request_args:dict, download_msg:str, crs:QgsCoordinateReferenceSystem, feedback_progress: bool, feedback: QgsProcessingFeedback, working_dir:str = None, output_dir:str = None) -> str:
    # set working and output directory
    working_dir, output_dir = self.setWorkAndOutputDir(working_dir, output_dir)
      
    # show download message, if specified
    if len(download_msg) > 0:
      feedback.pushInfo(download_msg)
    
    # request to the URL, with stream mode
    try:
      response = session.request(stream = True, timeout = 12.0, **request_args)
    except Exception as e:
      feedback.reportError(self.tr("Download was failed."))
      raise e
    
    # if the response is OK
    if response.status_code != 200:
      raise Exception(self.tr("Download was failed."))
    
    # get file size and prepare file path
    download_size = int(response.headers.get('content-length', 0))    
    download_path = os.path.join(working_dir, os.path.basename(request_args["url"]))

    # write downloaded file and show progress
    with open(download_path, 'wb') as file:
      for data in response.iter_content(chunk_size=1024):
        file.write(data)
        if feedback_progress:
          feedback.setProgress(int((file.tell() / download_size) * 100))
    
    # reset progress
    if feedback_progress:
      feedback.setProgress(0)      
      
    # unzip the downloaded archives
    tengun_paths = self.extractArchive(download_path, feedback)
    
    # assign projection to the tengun files and move them to the output folder
    for tng in tengun_paths:
      tng_dest_path = os.path.join(output_dir, os.path.basename(tng))
      self.assignTengunProjection(tng, crs, feedback, tng_dest_path)
      
      feedback.pushInfo(self.tr("Point cloud file was fetched: ") + tng_dest_path)
      feedback.pushInfo("")
    
    # delete temporary files
    os.remove(download_path)
    if any([download_path.endswith(ext) for ext in self.ARCHIVE_EXTENSION]):
      shutil.rmtree(os.path.join(os.path.dirname(download_path), os.path.splitext(os.path.basename(download_path))[0]))
      
    return None
  
  # extract the downloaded archives
  def extractArchive(self, file_path:str, feedback) -> list:    
    extracted_path = []
    # unzip the downloaded archives
    if any([file_path.endswith(ext) for ext in self.ARCHIVE_EXTENSION]):
      extract_path = os.path.join(os.path.dirname(file_path), os.path.splitext(os.path.basename(file_path))[0])
      feedback.pushInfo(self.tr("Unpacking archive ..."))
      shutil.unpack_archive(file_path, extract_path)
      for unpacked_file in glob.glob(os.path.join(extract_path, "*")):
        if any([unpacked_file.endswith(ext) for ext in self.TENGUN_EXTENSION]):
          extracted_path.append(unpacked_file)
    elif any([file_path.endswith(ext) for ext in self.TENGUN_EXTENSION]):
      extracted_path.append(file_path)
    
    return extracted_path
  
  def assignTengunProjection(self, tengun_path:str, crs:QgsCoordinateReferenceSystem, feedback: QgsProcessingFeedback, output_path:str=None) -> None:
    if output_path is None:
      output_path = tengun_path
      
    if crs.isValid():
      processing.run(
        "pdal:assignprojection", 
        {
          "INPUT": tengun_path,
          "CRS": crs,
          "OUTPUT": output_path
        },
        feedback = feedback,
        is_child_algorithm=True
      )["OUTPUT"]
    
    else:
      shutil.move(tengun_path, output_path)
      
    return None
  
  # download file(s) from the URL (asynchronously)
  def fetchMultipleTenguns(self, parameters: dict, context: QgsProcessingContext, feedback: QgsProcessingFeedback, parallel:bool = False, working_dir:str = None, output_dir:str = None) -> None:
    # set working and output directory
    working_dir, output_dir = self.setWorkAndOutputDir(working_dir, output_dir)
    
    if not self.WEBFETCH_ARGS["SET"]:
      return
    # if login is needed, session is also needed
    if self.WEBFETCH_ARGS.get("LOGIN") is None:
      session = requests.Session()
    else:
      session = self.WEBFETCH_ARGS["LOGIN"]["SESSION"]
      if not isinstance(session, requests.sessions.Session):
        feedback.reportError(self.tr("Login session is not given."))
        raise Exception(self.tr("Login session is not given"))
        
    # set download arguments
    n_url = len(self.WEBFETCH_ARGS["REQUEST_ARGS"])
    download_args = [
      {
        "download_msg": f"{i}/{n_url} Downloading {ra['url']}", 
        "request_args": {k: ra[k] for k in ra.keys() if k != "crs"},
        "crs": ra["crs"]
      } for i, ra in zip(range(1, n_url + 1), self.WEBFETCH_ARGS["REQUEST_ARGS"])
    ]
    mx_wk = None if parallel else 1
    
    # fetch in parallel
    with concurrent.futures.ThreadPoolExecutor(max_workers=mx_wk) as executor:
      for _ in executor.map(
        lambda d_arg : self.fetchSingleTengun(
          session, 
          request_args = d_arg["request_args"], 
          download_msg=d_arg["download_msg"], 
          crs = d_arg["crs"],
          feedback_progress = not parallel, feedback = feedback, 
          working_dir=working_dir, output_dir=output_dir
          ), 
        download_args
        ):
        if feedback.isCanceled():
          feedback.reportError(self.tr("Fetching was canceled."))
          raise Exception(self.tr("Fetching was canceled."))
  
    